# vi:shiftwidth=2:tabstop=2:expandtab:smartindent

# Inverse level design platformer test

import pygame, sys, math

state_unknown = 0
state_block = 1
state_empty = 2

state_grid = []

block_size = 32

grid_changed = 1

def update_grid(surface):
  global grid_changed
  if grid_changed:
    surface.fill((0,0,0))
    for x in range(0, width, block_size):
      for y in range(0, height, block_size):
        if state_grid[x/block_size][y/block_size] == state_block:
          surface.fill((182,182,182), ((x,y), (block_size,block_size)))
        if state_grid[x/block_size][y/block_size] == state_empty:
          surface.fill((100,100,100), ((x,y), (block_size,block_size)))
    grid_changed = 0

def can_enter(pos,dontchange=0):
  global grid_changed, state_grid
  state = state_grid[pos[0]/block_size][pos[1]/block_size]
  if state == state_unknown:
    if dontchange:
      # don't assume that we can enter if we can't change things to assure it
      #print "would have changed!"
      return 0
    state_grid[pos[0]/block_size][pos[1]/block_size] = state_empty
    grid_changed = 1
    return 1
  if state == state_block:
    return 0
  if state == state_empty:
    return 1

def can_avoid_entering(pos):
  global grid_changed, state_grid
  state = state_grid[pos[0]/block_size][pos[1]/block_size]
  if state == state_unknown:
    state_grid[pos[0]/block_size][pos[1]/block_size] = state_block
    grid_changed = 1
    return 1
  if state == state_block:
    return 1
  if state == state_empty:
    return 0

guy_size = 10

def guy_can_be_here(pos,dontchange=0):
  corner_1 = can_enter((pos[0]-guy_size,pos[1]-guy_size),dontchange)
  corner_2 = can_enter((pos[0]-guy_size,pos[1]+guy_size),dontchange)
  corner_3 = can_enter((pos[0]+guy_size,pos[1]-guy_size),dontchange)
  corner_4 = can_enter((pos[0]+guy_size,pos[1]+guy_size),dontchange)

  return corner_1 and corner_2 and corner_3 and corner_4

def guy_can_fall_here(pos):
  # Maybe we need non-destructive logic here, to only draw in a corner here
  # if there isn't one already?
  corner_1 = can_avoid_entering((pos[0]-guy_size, pos[1]+guy_size))
  corner_2 = can_avoid_entering((pos[0]+guy_size, pos[1]+guy_size))
  return not (corner_1 or corner_2)

def clip(oldpos, pos):
  # Starting at oldpos, step one pixel at a time until we hit an invalid pixel.
  # Slow and dumb, but I'm not up to thinking of a faster way.
  lastok_pos = oldpos
  cur_pos = lastok_pos
  diff = (pos[0]-oldpos[0], pos[1]-oldpos[1])
  diff_count = (abs(diff[0]), abs(diff[1]))
  count = 0
  if diff_count[0] > diff_count[1]:
    diff_sign = diff[0]/diff_count[0]
    while guy_can_be_here(cur_pos,1):
      lastok_pos = cur_pos
      cur_pos = (cur_pos[0]+diff_sign,oldpos[1]+diff[1]*count/diff_count[0])
      count = count + 1
  else:
    diff_sign = diff[1]/diff_count[1]
    while guy_can_be_here(cur_pos,1):
      lastok_pos = cur_pos
      cur_pos = (oldpos[0]+diff[0]*count/diff_count[1],cur_pos[1]+diff_sign)
      count = count + 1

#  print "oldpos =   ", oldpos
#  print "pos = ", pos
#  print "lastok_pos = ", lastok_pos

  return lastok_pos

pygame.init()

#screen_size = width, height = 1200, 900
screen_size = width, height = 800, 600

for blah in range(0, int(math.ceil(float(width)/block_size))):
  state_grid.append([state_unknown] * int(math.ceil(float(height)/block_size)))

for x in range(0, int(math.ceil(float(width)/block_size))):
  state_grid[x][0] = state_block
  state_grid[x][int(math.ceil(float(height)/block_size))-1] = state_block

for y in range(0, int(math.ceil(float(height)/block_size))):
  state_grid[0][y] = state_block
  state_grid[int(math.ceil(float(width)/block_size))-1][y] = state_block

pos = ( width/2*256, (height/block_size*block_size-guy_size-1)*256 )
vel = ( 0, 0 )
oldpos = pos
gravity_acc = 30
walk_acc = 50
jump_acc = 3*256
on_ground = 1
jump_recovered = 1

vel_limit = (15*walk_acc, 30*gravity_acc)

screen = pygame.display.set_mode(screen_size)
grid = pygame.Surface(screen_size)
grid = grid.convert()

clock = pygame.time.Clock()

pygame.event.set_allowed(pygame.QUIT)

#pygame.key.set_repeat(500,25)

while 1:
  # ensure that our current space is empty on first loop, used as sanity
  # check later
  if not guy_can_be_here((pos[0]/256,pos[1]/256)):
    print "You're not supposed to be here. Forget all this."
    print "pos:", pos
    print "vel:", vel
    print "on_ground:", on_ground
    exit()

  # draw
  if pos != oldpos or grid_changed:
    update_grid(grid)
    screen.blit(grid, (0,0))
#    pygame.draw.circle(screen, (255, 255, 255), (pos[0]/256,pos[1]/256), ball_radius)
    screen.fill((255,255,255), ((pos[0]/256-guy_size,pos[1]/256-guy_size), (guy_size*2,guy_size*2)))
    pygame.display.flip()

  oldpos = pos
  clock.tick_busy_loop(60)

  pygame.event.pump()

  for event in pygame.event.get():
    if event.type == pygame.QUIT:
      sys.exit()

  keys = pygame.key.get_pressed()

  if keys[pygame.K_ESCAPE]:
    sys.exit()

  # if we're on the ground we can think about horizontal acceleration
  if on_ground:
    # horizontal acceleration

    # accelerate right
    if keys[pygame.K_RIGHT]:
      if vel[0] < 0:
        # instant turnaround
        vel = (0, vel[1])
      else:
        vel = (vel[0] + walk_acc, vel[1])
        if vel[0] > vel_limit[0]:
          vel = (vel_limit[0], vel[1])

    # accelerate left
    if keys[pygame.K_LEFT]:
      if vel[0] > 0:
        # instant turnaround
        vel = (0, vel[1])
      else:
        vel = (vel[0] - walk_acc, vel[1])
        if vel[0] < -vel_limit[0]:
          vel = (-vel_limit[0], vel[1])

    # friction (if not moving, immediate stop)
    if (not keys[pygame.K_RIGHT]) and (not keys[pygame.K_LEFT]):
      vel = (0, vel[1])

  # jumping
  if keys[pygame.K_SPACE]:
    if on_ground:
      if jump_recovered:
        jump_recovered = 0
        vel = (vel[0], -jump_acc)
  else:
    if on_ground:
      # needs one cycle of not holding down spacebar while on ground
      jump_recovered = 1

  # gravity
  if not on_ground:
    vel = (vel[0], vel[1] + gravity_acc)
    if vel[1] > vel_limit[1]:
      vel = (vel[0], vel_limit[1])

  # try horizontal movement before doing fall logic
  if not guy_can_be_here(( (pos[0]+vel[0])/256, pos[1]/256)):
    vel = (0, vel[1])

  pos = (pos[0] + vel[0], pos[1] + vel[1])

  # falling?
  if vel[1] > 0:
    if not guy_can_fall_here((pos[0]/256,pos[1]/256)):
      on_ground = 1 # hit ground
    else:
      on_ground = 0 # didn't hit yet
  # rising
  if vel[1] < 0:
    on_ground = 0 # always above ground if we're moving up
  # standing over nothing?
  if vel[1] == 0:
    if guy_can_fall_here((pos[0]/256,pos[1]/256+1)):
      on_ground = 0

  # check for collision
  if not guy_can_be_here((pos[0]/256,pos[1]/256)):
    pos = clip((oldpos[0]/256,oldpos[1]/256), (pos[0]/256,pos[1]/256))
    pos = (pos[0]*256,pos[1]*256)
    vel = (vel[0], 0)

  if keys[pygame.K_F1]:
    print "pos:", pos
    print "vel:", vel
    print "on_ground:", on_ground
